/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2016 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.download;

import org.apache.http.HttpStatus;
import org.artifactory.api.archive.ArchiveType;
import org.artifactory.api.common.BasicStatusHolder;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.download.FolderDownloadInfo;
import org.artifactory.api.repo.exception.ItemNotFoundRuntimeException;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.api.storage.StorageUnit;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.download.FolderDownloadConfigDescriptor;
import org.artifactory.fs.ItemInfo;
import org.artifactory.repo.Repo;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.spring.Reloadable;
import org.artifactory.storage.fs.service.FileService;
import org.artifactory.version.CompoundVersionDetails;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.io.InputStream;
import java.util.concurrent.Semaphore;

import static org.apache.http.HttpStatus.*;

/**
 * Serves requests for download of folders as archive
 *
 * @author Dan Feldman
 */
@Service
@Reloadable(beanClass = InternalFolderDownloadService.class, initAfter = {InternalRepositoryService.class})
public class FolderDownloadServiceImpl implements InternalFolderDownloadService {
    private static final Logger log = LoggerFactory.getLogger(FolderDownloadServiceImpl.class);

    @Autowired
    private InternalRepositoryService repoService;

    @Autowired
    private FileService fileService;

    @Autowired
    private AuthorizationService authService;

    private int maxDownloadSizeMb;
    private long maxFiles;
    private boolean serviceEnabled;
    private ConcurrentDownloadCounter concurrentDownloadCounter;

    @Override
    public void init() {
        FolderDownloadConfigDescriptor config = getFolderDownloadConfig();
        this.maxDownloadSizeMb = config.getMaxDownloadSizeMb();
        this.maxFiles = config.getMaxFiles();
        this.serviceEnabled = config.isEnabled();
        this.concurrentDownloadCounter = new ConcurrentDownloadCounter(config.getMaxConcurrentRequests(), true);
    }

    @Override
    public void reload(CentralConfigDescriptor oldDescriptor) {
        FolderDownloadConfigDescriptor newConfig = getFolderDownloadConfig();
        FolderDownloadConfigDescriptor oldConfig = oldDescriptor.getFolderDownloadConfig();
        if (!newConfig.equals(oldConfig)) {
            this.maxDownloadSizeMb = newConfig.getMaxDownloadSizeMb();
            this.maxFiles = newConfig.getMaxFiles();
            this.serviceEnabled = newConfig.isEnabled();
            //Take the extra effort to compare old and new because resize is a synchronized method
            if (newConfig.getMaxConcurrentRequests() != oldConfig.getMaxConcurrentRequests()) {
                concurrentDownloadCounter.resize(newConfig.getMaxConcurrentRequests());
            }
        }
    }

    @Override
    public boolean getAvailableDownloadSlot() {
        if (!concurrentDownloadCounter.tryAcquire()) {
            log.debug("No available download slots, current available count in semaphore: {}",
                    concurrentDownloadCounter.availablePermits());
            return false;
        }
        return true;
    }

    @Override
    public void releaseDownloadSlot() {
        concurrentDownloadCounter.release();
    }

    @Override
    public InputStream process(RepoPath pathToDownload, ArchiveType archiveType, boolean includeChecksumFiles,
            BasicStatusHolder status) {
        if (!serviceEnabled) {
            status.error("Downloading folders as archive was disabled by your system admin.", SC_FORBIDDEN, log);
            return null;
        }
        FolderDownloadInfo info = collectFolderInfo(pathToDownload);
        if (pathExceedsLimits(info, status, pathToDownload.toPath())) {
            return null;
        }
        try {
            if (!getAvailableDownloadSlot()) {
                status.error("There are too many folder download requests currently running, try again later.",
                        HttpStatus.SC_BAD_REQUEST, log);
                return null;
            }
            return process(pathToDownload, archiveType, maxDownloadSizeMb, maxFiles, includeChecksumFiles, status);
        } catch (Exception e) {
            status.error("Error executing folder download: " + e.getMessage(), log);
            log.debug("Caught exception while creating output stream for folder download: ", e);
        }
        return null;
    }

    @Override
    public InputStream processNoLimit(RepoPath pathToDownload, ArchiveType archiveType, boolean includeChecksumFiles,
            BasicStatusHolder status) {
        if (!assertPath(pathToDownload, status)) {
            return null;
        }
        return process(pathToDownload, archiveType, 0, 0, includeChecksumFiles, status);
    }

    private InputStream process(RepoPath pathToDownload, ArchiveType archiveType, int maxDownloadSizeMb, long maxFiles,
                                boolean includeChecksumFiles, BasicStatusHolder status) {
        if (!assertPath(pathToDownload, status)) {
            return null;
        }
        FolderArchiveStreamer streamer = new FolderArchiveStreamer(pathToDownload, archiveType, maxDownloadSizeMb,
                maxFiles, includeChecksumFiles, status);
        return streamer.go();
    }

    @Override
    public FolderDownloadInfo collectFolderInfo(RepoPath folder) {
        return new FolderDownloadInfo(StorageUnit.MB.fromBytes(fileService.getFilesTotalSize(folder)),
                repoService.getArtifactCount(folder));
    }

    @Override
    public FolderDownloadConfigDescriptor getFolderDownloadConfig() {
        return ContextHelper.get().beanForType(CentralConfigService.class).getDescriptor().getFolderDownloadConfig();
    }

    private boolean assertPath(RepoPath pathToDownload, BasicStatusHolder status) {
        String folderPath = pathToDownload.toPath();
        ItemInfo itemInfo;
        Repo repo = repoService.repositoryByKey(pathToDownload.getRepoKey());
        if (repo == null) {
            status.error(pathToDownload.getRepoKey() + " is not a repository.", SC_NOT_FOUND, log);
            return false;
        } else if (!repo.isLocal() && !repo.isCache()) {
            status.error("Downloading a folder or a repository's root is only available for local (or cache) " +
                    "repositories", SC_NOT_FOUND, log);
            return false;
        }
        if (!pathToDownload.isRoot()) {
            //Not root, check folder status
            try {
                itemInfo = repoService.getItemInfo(pathToDownload);
            } catch (ItemNotFoundRuntimeException inf) {
                status.error("Path '" + folderPath + "' does not exist, aborting folder download", SC_NOT_FOUND, log);
                return false;
            }
            if (!itemInfo.isFolder()) {
                status.error("Path '" + folderPath + "' is not a folder, aborting folder download", SC_BAD_REQUEST, log);
                return false;
            }
        }
        if (!authService.canRead(pathToDownload)) {
            status.error("You don't have the required permissions to download " + folderPath + ".", SC_FORBIDDEN, log);
            return false;
        }
        return true;
    }

    private boolean pathExceedsLimits(FolderDownloadInfo info, BasicStatusHolder status, String folderPath) {
        if (info.getSizeMb() > maxDownloadSizeMb) {
            status.error("Size of path '" + folderPath + "' (" + String.format("%.2f", info.getSizeMb()) + "MB) exceeds"
                    + " the max allowed " + "folder download size (" + maxDownloadSizeMb + "MB).", SC_BAD_REQUEST, log);
            return true;
        } else if (info.getTotalFiles() > maxFiles) {
            status.error("Number of files under the path '" + folderPath + "' (" + info.getTotalFiles() + ") exceeds " +
                    "the max allowed file count for folder download (" + maxFiles + ").", SC_BAD_REQUEST, log);
            return true;
        }
        return false;
    }

    @Override
    public void destroy() {
    }

    @Override
    public void convert(CompoundVersionDetails source, CompoundVersionDetails target) {
    }

    /**
     * A resizable semaphore to act as the concurrent downloads counter for the service.
     */
    private static class ConcurrentDownloadCounter extends Semaphore {

        private int permits;

        public ConcurrentDownloadCounter(int permits, boolean fair) {
            super(permits, fair);
            this.permits = permits;
        }

        public synchronized void resize(int newSize) {
            log.debug("Resizing download counter, old size: {}, new size: {}", permits, newSize);
            int delta = newSize - permits;
            if (delta == 0) {
                log.trace("Same size chosen - no need to resize");
                return;
            } else if (delta > 0) {
                log.trace("Adding {} permits", delta);
                this.release(delta);
            } else if (delta < 0) {
                log.trace("Reducing {} permits", Math.abs(delta));
                this.reducePermits(Math.abs(delta));
            }
            this.permits = newSize;
            log.debug("Current available permits in counter: {}", this.availablePermits());
        }
    }
}
