/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2016 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.addon.build;

import org.artifactory.addon.Addon;
import org.artifactory.api.common.BasicStatusHolder;
import org.artifactory.api.rest.build.artifacts.BuildArtifactsRequest;
import org.artifactory.api.rest.build.diff.BuildsDiff;
import org.artifactory.build.ArtifactoryBuildArtifact;
import org.artifactory.fs.FileInfo;
import org.artifactory.sapi.common.Lock;
import org.jfrog.build.api.BaseBuildFileBean;
import org.jfrog.build.api.Build;
import org.jfrog.build.api.BuildRetention;
import org.jfrog.build.api.Dependency;
import org.jfrog.build.api.dependency.BuildPatternArtifacts;
import org.jfrog.build.api.dependency.BuildPatternArtifactsRequest;

import javax.annotation.Nonnull;
import java.io.File;
import java.io.IOException;
import java.util.Map;
import java.util.Set;

/**
 * Addon for continuous integration info handling
 *
 * @author Noam Y. Tenne
 */
public interface BuildAddon extends Addon {

    /**
     * Returns a mapping of file info to artifact for all artifacts in this build, will return artifacts that were not
     * directly involved in the build (but same checksum!) if the build's artifacts themselves were'nt found based on
     * their properties.
     * NOTE: a mapping might be null if the artifact does not exist in artifactory
     *
     * @param build The searched build
     * @return file beans and file infos
     */
    Set<ArtifactoryBuildArtifact> getBuildArtifactsFileInfosWithFallback(Build build);

    /**
     * Returns a mapping of file info to artifact for all artifacts in this build.
     * NOTE: a mapping might be null if the artifact does not exist in artifactory
     *
     * @param build The searched build
     * @return file beans and file infos
     */
    Set<ArtifactoryBuildArtifact> getBuildArtifactsFileInfos(Build build);

    /**
     * Returns a file info of dependency object for a build file bean
     * NOTE: a mapping might be null if the dependency does not exist in artifactory
     *
     * @param build The searched build
     * @return file beans and file infos
     */
    Map<Dependency, FileInfo> getBuildDependenciesFileInfos(Build build);

    /**
     * Searches for artifacts with build name properties that contain 'from' as a value and renames them to 'to'
     *
     * @param from Build name property value to search for
     * @param to   Replacement build name
     */
    @Lock
    void renameBuildNameProperty(String from, String to);

    /**
     * Discard old builds according to a maximum date. If a build is below the {@code minimumBuildDate} then that build
     * will be discarded.
     *
     * @param buildName         The name of the build.
     * @param buildRetention    Build retention model that holds information about which build to discard
     * @param multiStatusHolder Status holder
     */
    @Lock
    void discardOldBuildsByDate(String buildName, BuildRetention buildRetention, BasicStatusHolder multiStatusHolder);

    /**
     * Discard old builds according to the maximum amount of builds that should be retained. if {@code count} is larger
     * than the size of the set of builds, all builds will be retained.
     *
     * @param buildName         The name of the build.
     * @param discard           Build retention model that holds information about which build to discard
     * @param multiStatusHolder Status holder
     */
    @Lock
    void discardOldBuildsByCount(String buildName, BuildRetention discard, BasicStatusHolder multiStatusHolder);

    /**
     * Returns the built artifacts matching the request
     *
     * @param buildPatternArtifactsRequest contains build name and build number or keyword
     * @param servletContextUrl            for building urls of current Artifactory instance
     * @return artifacts generated by the build
     */
    BuildPatternArtifacts getBuildPatternArtifacts(@Nonnull BuildPatternArtifactsRequest buildPatternArtifactsRequest,
            String servletContextUrl);

    /**
     * Returns build artifacts map according to the param input regexp patterns.
     *
     * @param buildArtifactsRequest A wrapper which contains the necessary parameters
     * @return A map from {@link FileInfo}s to their target directories relative paths
     * @see BuildArtifactsRequest
     */
    Map<FileInfo, String> getBuildArtifacts(BuildArtifactsRequest buildArtifactsRequest);

    /**
     * Returns an archive file according to the param archive type (zip/tar/tar.gz/tgz) which contains
     * all build artifacts according to the given build name and number (can be latest or latest by status).
     *
     * @param buildArtifactsRequest A wrapper which contains the necessary parameters
     * @return The archived file of build artifacts with their hierarchy rules
     * @see BuildArtifactsRequest
     */
    File getBuildArtifactsArchive(BuildArtifactsRequest buildArtifactsRequest) throws IOException;

    /**
     * Returns diff object between two given builds (same build name, different numbers)
     *
     * @param firstBuild         The first build to compare, must be newer than the second build
     * @param secondBuild        The second build to compare against
     * @param baseStorageInfoUri Base storage uri
     */
    BuildsDiff getBuildsDiff(Build firstBuild, Build secondBuild, String baseStorageInfoUri);

    /**
     * return artifact or dependency file info (in order to get repo path)
     *
     * @param artifact - artifact or dependency  (to get file info for)
     * @param build    - build data
     * @return file info instance of artifact or dependency
     */
    FileInfo getFileBeanInfo(BaseBuildFileBean artifact, Build build);

}