/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2016 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.security.providermgr;

import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import org.apache.commons.lang.StringUtils;
import org.artifactory.common.ConstantValues;
import org.artifactory.security.props.auth.model.OauthDockerErrorModel;
import org.artifactory.security.props.auth.model.OauthErrorModel;
import org.artifactory.security.props.auth.model.OauthModel;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Component;

import javax.annotation.Nonnull;
import javax.annotation.PostConstruct;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

/**
 * artifactory specific token cache implementation ,
 * The tokens are cached in-memory and auto expire according to the
 *
 * @author Chen Keinan
 */
@Component
public class ArtifactoryTokenProvider {
    private static final Logger log = LoggerFactory.getLogger(ArtifactoryTokenProvider.class);

    private LoadingCache<ArtifactoryCacheKey, OauthModel> tokens;

    @PostConstruct
    public void initTokensCache() {
        tokens = CacheBuilder.newBuilder()
                .initialCapacity(1000)
                .expireAfterWrite(ConstantValues.genericTokensCacheIdleTimeSecs.getLong(), TimeUnit.SECONDS)
                .build(new CacheLoader<ArtifactoryCacheKey, OauthModel>() {
                    @Override
                    public OauthModel load(@Nonnull ArtifactoryCacheKey key) throws Exception {
                        return fetchNewToken(key);
                    }
                });
    }

    public OauthModel getToken(ArtifactoryCacheKey artifactoryCacheKey) {
        String userName = artifactoryCacheKey.getUser();
        OauthModel oauthModel = null;
        try {
            log.trace("Getting token for " + userName);
            oauthModel = tokens.get(artifactoryCacheKey);
            return oauthModel;
        } catch (ExecutionException e) {
            throw new RuntimeException("Could not get token from cache for " + userName, e);
        } finally {
            if (oauthModel != null) {
                if (oauthModel instanceof OauthErrorModel || oauthModel instanceof OauthDockerErrorModel) {
                    tokens.invalidate(artifactoryCacheKey);
                }
            }
        }
    }

    /**
     * Used when expiring user credentials or revoking api keys - will remove the user's tokens from the cache
     * to force re-authentication.
     * @param userName  user to invalidate cache entries for
     */
    public void invalidateUserCacheEntries(String userName) {
        if(StringUtils.isNotBlank(userName)) {
            List<ArtifactoryCacheKey> toInvalidate = tokens.asMap().keySet().stream()
                    .filter(Objects::nonNull)
                    .filter(cacheKey -> userName.equalsIgnoreCase(cacheKey.getUser()))
                    .collect(Collectors.toList());
            tokens.invalidateAll(toInvalidate);
        }
    }

    public void invalidateCacheEntriesForAllUsers() {
        tokens.invalidateAll();
    }

    private OauthModel fetchNewToken(ArtifactoryCacheKey artifactoryCacheKey) {
        return artifactoryCacheKey.getProviderMgr().fetchTokenFromProvider();
    }
}
