/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2016 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.storage.db.aql.service.decorator;

import com.google.common.collect.Lists;
import org.artifactory.aql.model.AqlComparatorEnum;
import org.artifactory.aql.model.AqlField;
import org.artifactory.aql.model.AqlPhysicalFieldEnum;
import org.artifactory.aql.model.AqlRepoProvider;
import org.artifactory.aql.model.AqlValue;
import org.artifactory.aql.model.AqlVariable;
import org.artifactory.aql.model.AqlVariableTypeEnum;
import org.artifactory.storage.db.aql.sql.builder.query.aql.AqlAdapter;
import org.artifactory.storage.db.aql.sql.builder.query.aql.AqlQuery;
import org.artifactory.storage.db.aql.sql.builder.query.aql.AqlQueryElement;
import org.artifactory.storage.db.aql.sql.builder.query.aql.Criteria;
import org.artifactory.storage.db.aql.sql.builder.query.aql.OperatorQueryElement;
import org.artifactory.storage.db.aql.sql.builder.query.aql.SimpleCriteria;

import java.util.List;
import java.util.function.Predicate;
import java.util.stream.Collectors;

/**
 * <p>Created on 29/05/16
 *
 * @author Yinon Avraham
 */
public class VirtualRepoCriteriaDecorator implements DecorationStrategy {

    @Override
    public void decorate(AqlQuery aqlQuery, AqlQueryDecoratorContext decoratorContext) {
        List<AqlQueryElement> aqlQueryElements = aqlQuery.getAqlElements();
        int index = 0;
        while ((index = findNextRepoCriteria(index, aqlQueryElements)) >= 0) {
            index += handleRepoCriteria(index, aqlQueryElements, decoratorContext);
        }
    }

    private int findNextRepoCriteria(int start, List<AqlQueryElement> aqlQueryElements) {
        int index = start;
        for (; index < aqlQueryElements.size(); index++) {
            AqlQueryElement element = aqlQueryElements.get(index);
            if (element instanceof Criteria) {
                Criteria criteria = (Criteria) element;
                AqlVariable var1 = criteria.getVariable1();
                if (var1 instanceof AqlField && ((AqlField)var1).getFieldEnum() == AqlPhysicalFieldEnum.itemRepo) {
                    return index;
                }
            }
        }
        return -1;
    }

    private int handleRepoCriteria(int index, List<AqlQueryElement> aqlQueryElements, AqlQueryDecoratorContext decoratorContext) {
        Criteria criteria = (Criteria) aqlQueryElements.get(index);
        AqlComparatorEnum comparator = AqlComparatorEnum.value(criteria.getComparatorName());
        switch (comparator) {
            case equals:
            case matches:
            case greater:
            case greaterEquals:
            case less:
            case lessEquals:
                return handlePositive(index, aqlQueryElements, criteria, comparator, decoratorContext);
            case notEquals:
            case notMatches:
                return handleNegative(index, aqlQueryElements, criteria, comparator, decoratorContext);
        }
        throw new IllegalArgumentException("Unexpected repo comparator: " + comparator.signature);
    }

    private int handlePositive(int index, List<AqlQueryElement> aqlQueryElements, Criteria criteria,
            AqlComparatorEnum comparator, AqlQueryDecoratorContext decoratorContext) {
        String value = (String) ((AqlValue)criteria.getVariable2()).toObject();
        Predicate<String> repoKeyPredicate = createRepoKeyPredicate(comparator, value);
        boolean includeOriginalCriteria = comparator != AqlComparatorEnum.equals;
        List<String> repoKeys = findVirtualReposAndResolveRepoKeys(repoKeyPredicate, decoratorContext);
        return adjustQueryElements(index, aqlQueryElements, criteria, AqlComparatorEnum.equals, repoKeys,
                AqlAdapter.or, includeOriginalCriteria);
    }

    private int handleNegative(int index, List<AqlQueryElement> aqlQueryElements, Criteria criteria,
            AqlComparatorEnum comparator, AqlQueryDecoratorContext decoratorContext) {
        String value = (String) ((AqlValue)criteria.getVariable2()).toObject();
        Predicate<String> repoKeyPredicate = createRepoKeyPredicate(comparator, value);
        boolean includeOriginalCriteria = comparator != AqlComparatorEnum.notEquals;
        List<String> repoKeys = findVirtualReposAndResolveRepoKeys(repoKeyPredicate, decoratorContext);
        return adjustQueryElements(index, aqlQueryElements, criteria, AqlComparatorEnum.notEquals, repoKeys,
                AqlAdapter.and, includeOriginalCriteria);
    }

    private List<String> findVirtualReposAndResolveRepoKeys(Predicate<String> repoKeyPredicate,
            AqlQueryDecoratorContext decoratorContext) {
        AqlRepoProvider repoProvider = decoratorContext.getRepoProvider();
        return repoProvider.getVirtualRepoKeys().stream()
                .filter(repoKeyPredicate)
                .flatMap(virtualRepoKey -> repoProvider.getVirtualResolvedLocalAndCacheRepoKeys(virtualRepoKey).stream())
                .distinct()
                .collect(Collectors.toList());
    }

    private Predicate<String> createRepoKeyPredicate(AqlComparatorEnum comparator, String value) {
        switch (comparator) {
            case equals:
            case notEquals:
                return key -> key.equals(value);
            case matches:
            case notMatches:
                AqlMatchesPattern matchesPattern = AqlMatchesPattern.compile(value);
                return matchesPattern::matches;
            case greater:
                return key -> key.compareTo(value) > 0;
            case greaterEquals:
                return key -> key.compareTo(value) >= 0;
            case less:
                return key -> key.compareTo(value) < 0;
            case lessEquals:
                return key -> key.compareTo(value) <= 0;
        }
        throw new IllegalArgumentException("Unexpected comparator: " + comparator);
    }

    private int adjustQueryElements(int index, List<AqlQueryElement> aqlQueryElements, Criteria criteria,
            AqlComparatorEnum comparator, List<String> repoKeys, OperatorQueryElement operator, boolean includeOriginal) {
        if (repoKeys.size() > 0) {
            List<AqlQueryElement> newCompoundCriteria = newCompoundCriteria(criteria, comparator, repoKeys, operator, includeOriginal);
            aqlQueryElements.remove(index);
            aqlQueryElements.addAll(index, newCompoundCriteria);
            return index + newCompoundCriteria.size();
        }
        return index + 1;
    }

    private List<AqlQueryElement> newCompoundCriteria(Criteria criteria, AqlComparatorEnum comparator,
            List<String> repoKeys, OperatorQueryElement operator, boolean includeOriginal) {
        List<AqlQueryElement> newCompoundCriteria = Lists.newArrayList();
        boolean first = true;
        if (includeOriginal) {
            newCompoundCriteria.add(criteria);
            first = false;
        }
        for (String repoKey : repoKeys) {
            if (!first) {
                newCompoundCriteria.add(operator);
            } else {
                first = false;
            }
            newCompoundCriteria.add(newCriteriaFrom(criteria, comparator, repoKey));
        }
        if (newCompoundCriteria.size() > 1) {
            newCompoundCriteria.add(0, AqlAdapter.open);
            newCompoundCriteria.add(AqlAdapter.close);
        }
        return newCompoundCriteria;
    }

    private Criteria newCriteriaFrom(Criteria criteria, AqlComparatorEnum comparatorEnum, String repoKey) {
        AqlValue newValue = new AqlValue(AqlVariableTypeEnum.string, repoKey);
        return new SimpleCriteria(criteria.getSubDomains(), criteria.getVariable1(), criteria.getTable1(),
                comparatorEnum.signature, newValue, criteria.getTable2(), false);
    }
}
