/*
 *
 * Copyright 2016 JFrog Ltd. All rights reserved.
 * JFROG PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package org.artifactory.rest.common.service.admin.xray;

import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.xray.XrayAddon;
import org.artifactory.addon.xray.XrayRepo;
import org.artifactory.aql.AqlService;
import org.artifactory.aql.api.AqlApiElement;
import org.artifactory.aql.api.domain.sensitive.AqlApiItem;
import org.artifactory.aql.api.internal.AqlBase;
import org.artifactory.aql.model.AqlItemTypeEnum;
import org.artifactory.aql.result.AqlEagerResult;
import org.artifactory.aql.result.rows.AqlBaseFullRowImpl;
import org.artifactory.aql.result.rows.AqlItem;
import org.artifactory.rest.common.model.xray.XrayIndexStatsModel;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.rest.common.service.RestService;
import org.artifactory.security.AuthenticationHelper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import javax.annotation.Nonnull;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.property;
import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.type;
import static org.artifactory.aql.api.internal.AqlBase.and;
import static org.artifactory.aql.api.internal.AqlBase.or;

/**
 * @author Yinon Avraham
 */
@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
public class GetXrayIndexStatsService implements RestService {
    private static final Logger log = LoggerFactory.getLogger(GetXrayIndexStatsService.class);

    private static final String PROP_XRAY_INDEX_STARTED = "xray.index.started";
    private static final String PROP_XRAY_INDEX_COMPLETED = "xray.index.completed";

    @Autowired
    private AddonsManager addonsManager;

    @Autowired
    private AqlService aqlService;

    @Override
    public void execute(ArtifactoryRestRequest request, RestResponse response) {
        String remoteAddress = AuthenticationHelper.getRemoteAddress(AuthenticationHelper.getAuthentication());
        String err = "Failing Xray index request received from " + remoteAddress + ": ";
        XrayAddon xrayAddon = addonsManager.addonByType(XrayAddon.class);
        if (!xrayAddon.isXrayEnabled()) {
            String disabledErr = "Xray is disabled or not supported by this instance's license.";
            response.error(disabledErr).responseCode(400);
            log.debug(err + disabledErr);
            return;
        }

        XrayIndexStatsModel indexStats = getXrayIndexStats(xrayAddon);
        response.iModel(indexStats);
    }

    private XrayIndexStatsModel getXrayIndexStats(XrayAddon xrayAddon) {
        Map<XrayRepoWrapper, String> xrayRepo2AqlRepoKey = getEnabledXrayRepos(xrayAddon);
        Map<String, XrayRepoWrapper> aqlRepoKey2XrayRepo = getReversedMap(xrayRepo2AqlRepoKey);
        List<XrayRepo> xrayRepos = xrayRepo2AqlRepoKey.keySet().stream()
                .map(XrayRepoWrapper::getXrayRepo)
                .collect(Collectors.toList());
        XrayIndexStatsModel indexStats = new XrayIndexStatsModel(xrayRepos);
        if (!xrayRepo2AqlRepoKey.isEmpty()) {
            AqlEagerResult<AqlItem> queryResult = findFilesWithXrayProps(aqlRepoKey2XrayRepo.keySet());
            collectIndexStats(indexStats, queryResult, aqlRepoKey2XrayRepo);
        }
        return indexStats;
    }

    /**
     * Find all files the x-ray enabled repos with x-ray index status started/completed properties
     * @param repos the repos in which to search for the items
     * @return the AQL result
     */
    private AqlEagerResult<AqlItem> findFilesWithXrayProps(Collection<String> repos) {
        AqlApiElement<AqlApiItem> enabledRepos = createXrayEnabledReposClause(repos);
        AqlApiItem query = AqlApiItem.create()
                .filter(
                        and(
                                enabledRepos,
                                type().equal(AqlItemTypeEnum.file.signature),
                                or(
                                        property().key().equal(PROP_XRAY_INDEX_STARTED),
                                        property().key().equal(PROP_XRAY_INDEX_COMPLETED)
                                )
                        )
                )
                .include(AqlApiItem.property().key());
        return aqlService.executeQueryEager(query);
    }

    private void collectIndexStats(XrayIndexStatsModel indexStats, AqlEagerResult<AqlItem> queryResult,
            Map<String, XrayRepoWrapper> aqlRepoKey2xrayRepo) {
        queryResult.getResults().stream().forEach(item -> {
            String repoKey = aqlRepoKey2xrayRepo.get(item.getRepo()).getXrayRepo().getName();
            String propKey = ((AqlBaseFullRowImpl) item).getKey();
            switch (propKey) {
                case PROP_XRAY_INDEX_STARTED:
                    indexStats.increaseStarted(repoKey);
                    break;
                case PROP_XRAY_INDEX_COMPLETED:
                    indexStats.increaseCompleted(repoKey);
                    break;
            }
        });
    }

    /**
     * Get a map of x-ray repo mapped to the repo key to be used in the AQL filter.
     * I.e. a local repo is mapped to itself, a remote repo is mapped to its local cache repo.
     */
    private Map<XrayRepoWrapper, String> getEnabledXrayRepos(XrayAddon xrayAddon) {
        List<XrayRepo> xrayRepos = xrayAddon.getXrayIndexedAndNonIndexed(true);
        return xrayRepos.stream()
                .collect(Collectors.toMap(
                        xrayRepo -> new XrayRepoWrapper(xrayRepo),
                        xrayRepo -> "remote".equals(xrayRepo.getType()) ?
                                xrayRepo.getName() + "-cache" :
                                xrayRepo.getName()));
    }

    private AqlApiElement<AqlApiItem> createXrayEnabledReposClause(Collection<String> repos) {
        if (repos.size() == 1) {
            return createCriteria(repos.iterator().next());
        } else {
            AqlBase.OrClause<AqlApiItem> or = or();
            repos.stream().
                    map(GetXrayIndexStatsService::createCriteria).forEach(or::append);
            return or;
        }
    }

    public static AqlBase.CriteriaClause<AqlApiItem> createCriteria(String repoKey) {
        return AqlApiItem.repo().equal(repoKey);
    }

    /**
     * Get a reversed copy of the given map: value -> key.
     * Assumes the map is 1:1.
     */
    private <K, V> Map<V, K> getReversedMap(Map<K, V> map) {
        return map.entrySet().stream()
                .collect(Collectors.toMap(Map.Entry::getValue, Map.Entry::getKey));
    }

    private static class XrayRepoWrapper {
        private final XrayRepo xrayRepo;

        public XrayRepoWrapper(@Nonnull XrayRepo xrayRepo) {
            this.xrayRepo = xrayRepo;
        }

        public XrayRepo getXrayRepo() {
            return xrayRepo;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            XrayRepoWrapper that = (XrayRepoWrapper) o;

            return xrayRepo.getName().equals(that.xrayRepo.getName());

        }

        @Override
        public int hashCode() {
            return xrayRepo.getName().hashCode();
        }
    }
}